/*  Copyright 2010, 2014 Lawrence Sebald

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
*/

#ifndef YabauseGLView_h
#define YabauseGLView_h

#import <Cocoa/Cocoa.h>
#include <OpenGL/OpenGL.h>
#include <OpenGL/GL.h>

@interface YabauseGLView : NSOpenGLView {
    IBOutlet NSWindow *window;
    NSRect oldFrame;
    NSWindow *fsWindow;
    BOOL _isFullscreen;
}

- (id)initWithFrame:(NSRect)frameRect;

- (void)toggleFullscreen;

- (BOOL)acceptsFirstResponder;
- (void)keyDown:(NSEvent *)event;
- (void)keyUp:(NSEvent *)event;

- (void)showWindow;

- (void)reshape;
- (void)drawRect:(NSRect)rect;

@end

#endif /* !YabauseGLView_h */
